@@ -1,42 +1,47 @@
 'use strict';
+const Denque = require('denque');
 const EventEmitter = require('events');
 const ServerDescription = require('./server_description').ServerDescription;
-const ServerType = require('./server_description').ServerType;
+const ServerType = require('./common').ServerType;
 const TopologyDescription = require('./topology_description').TopologyDescription;
-const TopologyType = require('./topology_description').TopologyType;
-const monitoring = require('./monitoring');
-const calculateDurationInMs = require('../utils').calculateDurationInMs;
-const MongoTimeoutError = require('../error').MongoTimeoutError;
-const Server = require('./server');
+const TopologyType = require('./common').TopologyType;
+const events = require('./events');
+const Server = require('./server').Server;
 const relayEvents = require('../utils').relayEvents;
 const ReadPreference = require('../topologies/read_preference');
-const readPreferenceServerSelector = require('./server_selectors').readPreferenceServerSelector;
-const writableServerSelector = require('./server_selectors').writableServerSelector;
 const isRetryableWritesSupported = require('../topologies/shared').isRetryableWritesSupported;
 const CoreCursor = require('../cursor').CoreCursor;
 const deprecate = require('util').deprecate;
 const BSON = require('../connection/utils').retrieveBSON();
 const createCompressionInfo = require('../topologies/shared').createCompressionInfo;
 const isRetryableError = require('../error').isRetryableError;
-const isSDAMUnrecoverableError = require('../error').isSDAMUnrecoverableError;
+const isNodeShuttingDownError = require('../error').isNodeShuttingDownError;
+const maxWireVersion = require('../utils').maxWireVersion;
 const ClientSession = require('../sessions').ClientSession;
-const createClientInfo = require('../topologies/shared').createClientInfo;
 const MongoError = require('../error').MongoError;
+const MongoServerSelectionError = require('../error').MongoServerSelectionError;
 const resolveClusterTime = require('../topologies/shared').resolveClusterTime;
 const SrvPoller = require('./srv_polling').SrvPoller;
 const getMMAPError = require('../topologies/shared').getMMAPError;
+const makeStateMachine = require('../utils').makeStateMachine;
+const eachAsync = require('../utils').eachAsync;
+const emitDeprecationWarning = require('../../utils').emitDeprecationWarning;
+const ServerSessionPool = require('../sessions').ServerSessionPool;
+const makeClientMetadata = require('../utils').makeClientMetadata;
+const CMAP_EVENT_NAMES = require('../../cmap/events').CMAP_EVENT_NAMES;
+
+const common = require('./common');
+const drainTimerQueue = common.drainTimerQueue;
+const clearAndRemoveTimerFrom = common.clearAndRemoveTimerFrom;
+
+const serverSelection = require('./server_selection');
+const readPreferenceServerSelector = serverSelection.readPreferenceServerSelector;
+const writableServerSelector = serverSelection.writableServerSelector;
+// const selectServers = serverSelection.selectServers;
 
 // Global state
 let globalTopologyCounter = 0;
 
-// Constants
-const TOPOLOGY_DEFAULTS = {
-  localThresholdMS: 15,
-  serverSelectionTimeoutMS: 30000,
-  heartbeatFrequencyMS: 10000,
-  minHeartbeatFrequencyMS: 500
-};
-
 // events that we relay to the `Topology`
 const SERVER_RELAY_EVENTS = [
   'serverHeartbeatStarted',
@@ -48,17 +53,32 @@ const SERVER_RELAY_EVENTS = [
 
   // NOTE: Legacy events
   'monitoring'
-];
+].concat(CMAP_EVENT_NAMES);
 
 // all events we listen to from `Server` instances
-const LOCAL_SERVER_EVENTS = SERVER_RELAY_EVENTS.concat([
-  'error',
-  'connect',
-  'descriptionReceived',
-  'close',
-  'ended'
+const LOCAL_SERVER_EVENTS = ['error', 'connect', 'descriptionReceived', 'close', 'ended'];
+
+const STATE_CLOSING = common.STATE_CLOSING;
+const STATE_CLOSED = common.STATE_CLOSED;
+const STATE_CONNECTING = common.STATE_CONNECTING;
+const STATE_CONNECTED = common.STATE_CONNECTED;
+const stateTransition = makeStateMachine({
+  [STATE_CLOSED]: [STATE_CLOSED, STATE_CONNECTING],
+  [STATE_CONNECTING]: [STATE_CONNECTING, STATE_CLOSING, STATE_CONNECTED, STATE_CLOSED],
+  [STATE_CONNECTED]: [STATE_CONNECTED, STATE_CLOSING, STATE_CLOSED],
+  [STATE_CLOSING]: [STATE_CLOSING, STATE_CLOSED]
+});
+
+const DEPRECATED_OPTIONS = new Set([
+  'autoReconnect',
+  'reconnectTries',
+  'reconnectInterval',
+  'bufferMaxEntries'
 ]);
 
+const kCancelled = Symbol('cancelled');
+const kWaitQueue = Symbol('waitQueue');
+
 /**
  * A container of server instances representing a connection to a MongoDB topology.
  *
@@ -99,7 +119,22 @@ class Topology extends EventEmitter {
       seedlist = parseStringSeedlist(seedlist);
     }
 
-    options = Object.assign({}, TOPOLOGY_DEFAULTS, options);
+    options = Object.assign({}, common.TOPOLOGY_DEFAULTS, options);
+    options = Object.freeze(
+      Object.assign(options, {
+        metadata: makeClientMetadata(options),
+        compression: { compressors: createCompressionInfo(options) }
+      })
+    );
+
+    DEPRECATED_OPTIONS.forEach(optionName => {
+      if (options[optionName]) {
+        emitDeprecationWarning(
+          `The option \`${optionName}\` is incompatible with the unified topology, please read more by visiting http://bit.ly/2D8WfT6`,
+          'DeprecationWarning'
+        );
+      }
+    });
 
     const topologyType = topologyTypeFromSeedlist(seedlist, options);
     const topologyId = globalTopologyCounter++;
@@ -110,6 +145,7 @@ class Topology extends EventEmitter {
       return result;
     }, new Map());
 
+    this[kWaitQueue] = new Denque();
     this.s = {
       // the id of this topology
       id: topologyId,
@@ -117,6 +153,8 @@ class Topology extends EventEmitter {
       options,
       // initial seedlist of servers to connect to
       seedlist: seedlist,
+      // initial state
+      state: STATE_CLOSED,
       // the topology description
       description: new TopologyDescription(
         topologyType,
@@ -129,33 +167,43 @@ class Topology extends EventEmitter {
       ),
       serverSelectionTimeoutMS: options.serverSelectionTimeoutMS,
       heartbeatFrequencyMS: options.heartbeatFrequencyMS,
-      minHeartbeatIntervalMS: options.minHeartbeatIntervalMS,
+      minHeartbeatFrequencyMS: options.minHeartbeatFrequencyMS,
       // allow users to override the cursor factory
       Cursor: options.cursorFactory || CoreCursor,
       // the bson parser
-      bson: options.bson || new BSON(),
+      bson:
+        options.bson ||
+        new BSON([
+          BSON.Binary,
+          BSON.Code,
+          BSON.DBRef,
+          BSON.Decimal128,
+          BSON.Double,
+          BSON.Int32,
+          BSON.Long,
+          BSON.Map,
+          BSON.MaxKey,
+          BSON.MinKey,
+          BSON.ObjectId,
+          BSON.BSONRegExp,
+          BSON.Symbol,
+          BSON.Timestamp
+        ]),
       // a map of server instances to normalized addresses
       servers: new Map(),
       // Server Session Pool
-      sessionPool: null,
+      sessionPool: new ServerSessionPool(this),
       // Active client sessions
-      sessions: [],
+      sessions: new Set(),
       // Promise library
       promiseLibrary: options.promiseLibrary || Promise,
       credentials: options.credentials,
       clusterTime: null,
 
       // timer management
-      monitorTimers: [],
-      iterationTimers: []
+      connectionTimers: new Set()
     };
 
-    // amend options for server instance creation
-    this.s.options.compression = { compressors: createCompressionInfo(options) };
-
-    // add client info
-    this.s.clientInfo = createClientInfo(options);
-
     if (options.srvHost) {
       this.s.srvPoller =
         options.srvPoller ||
@@ -178,6 +226,9 @@ class Topology extends EventEmitter {
 
       this.on('topologyDescriptionChanged', this.s.detectTopologyDescriptionChange);
     }
+
+    // NOTE: remove this when NODE-1709 is resolved
+    this.setMaxListeners(Infinity);
   }
 
   /**
@@ -191,17 +242,6 @@ class Topology extends EventEmitter {
     return BSON.native ? 'c++' : 'js';
   }
 
-  /**
-   * All raw connections
-   * @method
-   * @return {Connection[]}
-   */
-  connections() {
-    return Array.from(this.s.servers.values()).reduce((result, server) => {
-      return result.concat(server.s.pool.allConnections());
-    }, []);
-  }
-
   /**
    * Initiate server connect
    *
@@ -212,33 +252,40 @@ class Topology extends EventEmitter {
   connect(options, callback) {
     if (typeof options === 'function') (callback = options), (options = {});
     options = options || {};
+    if (this.s.state === STATE_CONNECTED) {
+      if (typeof callback === 'function') {
+        callback();
+      }
+
+      return;
+    }
+
+    stateTransition(this, STATE_CONNECTING);
 
     // emit SDAM monitoring events
-    this.emit('topologyOpening', new monitoring.TopologyOpeningEvent(this.s.id));
+    this.emit('topologyOpening', new events.TopologyOpeningEvent(this.s.id));
 
     // emit an event for the topology change
     this.emit(
       'topologyDescriptionChanged',
-      new monitoring.TopologyDescriptionChangedEvent(
+      new events.TopologyDescriptionChangedEvent(
         this.s.id,
         new TopologyDescription(TopologyType.Unknown), // initial is always Unknown
         this.s.description
       )
     );
 
+    // connect all known servers, then attempt server selection to connect
     connectServers(this, Array.from(this.s.description.servers.values()));
-    this.s.connected = true;
-
-    // otherwise, wait for a server to properly connect based on user provided read preference,
-    // or primary.
 
     translateReadPreference(options);
     const readPreference = options.readPreference || ReadPreference.primary;
-
     this.selectServer(readPreferenceServerSelector(readPreference), options, (err, server) => {
       if (err) {
+        this.close();
+
         if (typeof callback === 'function') {
-          callback(err, null);
+          callback(err);
         } else {
           this.emit('error', err);
         }
@@ -247,11 +294,13 @@ class Topology extends EventEmitter {
       }
 
       const errorHandler = err => {
+        stateTransition(this, STATE_CLOSED);
         server.removeListener('connect', connectHandler);
         if (typeof callback === 'function') callback(err, null);
       };
 
       const connectHandler = (_, err) => {
+        stateTransition(this, STATE_CONNECTED);
         server.removeListener('error', errorHandler);
         this.emit('open', err, this);
         this.emit('connect', this);
@@ -259,7 +308,6 @@ class Topology extends EventEmitter {
         if (typeof callback === 'function') callback(err, this);
       };
 
-      const STATE_CONNECTING = 1;
       if (server.s.state === STATE_CONNECTING) {
         server.once('error', errorHandler);
         server.once('connect', connectHandler);
@@ -274,21 +322,27 @@ class Topology extends EventEmitter {
    * Close this topology
    */
   close(options, callback) {
-    if (typeof options === 'function') (callback = options), (options = {});
-    options = options || {};
-
-    // clear all existing monitor timers
-    this.s.monitorTimers.map(timer => clearTimeout(timer));
-    this.s.monitorTimers = [];
-
-    this.s.iterationTimers.map(timer => clearTimeout(timer));
-    this.s.iterationTimers = [];
-
-    if (this.s.sessionPool) {
-      this.s.sessions.forEach(session => session.endSession());
-      this.s.sessionPool.endAllPooledSessions();
+    if (typeof options === 'function') {
+      callback = options;
+      options = {};
     }
 
+    if (typeof options === 'boolean') {
+      options = { force: options };
+    }
+
+    options = options || {};
+    if (this.s.state === STATE_CLOSED) {
+      if (typeof callback === 'function') {
+        callback();
+      }
+
+      return;
+    }
+
+    drainWaitQueue(this[kWaitQueue], new MongoError('Topology closed'));
+    drainTimerQueue(this.s.connectionTimers);
+
     if (this.s.srvPoller) {
       this.s.srvPoller.stop();
       if (this.s.handleSrvPolling) {
@@ -302,32 +356,28 @@ class Topology extends EventEmitter {
       delete this.s.detectTopologyDescriptionChange;
     }
 
-    const servers = this.s.servers;
-    if (servers.size === 0) {
-      this.s.connected = false;
-      if (typeof callback === 'function') {
-        callback(null, null);
-      }
+    stateTransition(this, STATE_CLOSING);
 
-      return;
-    }
+    this.s.sessions.forEach(session => session.endSession());
+    this.s.sessionPool.endAllPooledSessions(() => {
+      eachAsync(
+        Array.from(this.s.servers.values()),
+        (server, cb) => destroyServer(server, this, options, cb),
+        err => {
+          this.s.servers.clear();
 
-    // destroy all child servers
-    let destroyed = 0;
-    servers.forEach(server =>
-      destroyServer(server, this, () => {
-        destroyed++;
-        if (destroyed === servers.size) {
           // emit an event for close
-          this.emit('topologyClosed', new monitoring.TopologyClosedEvent(this.s.id));
+          this.emit('topologyClosed', new events.TopologyClosedEvent(this.s.id));
+
+          stateTransition(this, STATE_CLOSED);
+          this.emit('close');
 
-          this.s.connected = false;
           if (typeof callback === 'function') {
-            callback(null, null);
+            callback(err);
           }
         }
-      })
-    );
+      );
+    });
   }
 
   /**
@@ -370,30 +420,43 @@ class Topology extends EventEmitter {
     const transaction = session && session.transaction;
 
     if (isSharded && transaction && transaction.server) {
-      callback(null, transaction.server);
+      callback(undefined, transaction.server);
       return;
     }
 
-    // clear out any existing iteration timers
-    this.s.iterationTimers.map(timer => clearTimeout(timer));
-    this.s.iterationTimers = [];
+    // support server selection by options with readPreference
+    let serverSelector = selector;
+    if (typeof selector === 'object') {
+      const readPreference = selector.readPreference
+        ? selector.readPreference
+        : ReadPreference.primary;
 
-    selectServers(
-      this,
-      selector,
-      options.serverSelectionTimeoutMS,
-      process.hrtime(),
-      (err, servers) => {
-        if (err) return callback(err, null);
+      serverSelector = readPreferenceServerSelector(readPreference);
+    }
 
-        const selectedServer = randomSelection(servers);
-        if (isSharded && transaction && transaction.isActive) {
-          transaction.pinServer(selectedServer);
-        }
+    const waitQueueMember = {
+      serverSelector,
+      transaction,
+      callback
+    };
 
-        callback(null, selectedServer);
-      }
-    );
+    const serverSelectionTimeoutMS = options.serverSelectionTimeoutMS;
+    if (serverSelectionTimeoutMS) {
+      waitQueueMember.timer = setTimeout(() => {
+        waitQueueMember[kCancelled] = true;
+        waitQueueMember.timer = undefined;
+        const timeoutError = new MongoServerSelectionError(
+          `Server selection timed out after ${serverSelectionTimeoutMS} ms`,
+          this.description
+        );
+
+        waitQueueMember.callback(timeoutError);
+      }, serverSelectionTimeoutMS);
+    }
+
+    // place the member at the front of the wait queue
+    this[kWaitQueue].unshift(waitQueueMember);
+    processWaitQueue(this);
   }
 
   // Sessions related methods
@@ -402,10 +465,11 @@ class Topology extends EventEmitter {
    * @return Whether the topology should initiate selection to determine session support
    */
   shouldCheckForSessionSupport() {
-    return (
-      (this.description.type === TopologyType.Single && !this.description.hasKnownServers) ||
-      !this.description.hasDataBearingServers
-    );
+    if (this.description.type === TopologyType.Single) {
+      return !this.description.hasKnownServers;
+    }
+
+    return !this.description.hasDataBearingServers;
   }
 
   /**
@@ -421,10 +485,10 @@ class Topology extends EventEmitter {
   startSession(options, clientOptions) {
     const session = new ClientSession(this, this.s.sessionPool, options, clientOptions);
     session.once('ended', () => {
-      this.s.sessions = this.s.sessions.filter(s => !s.equals(session));
+      this.s.sessions.delete(session);
     });
 
-    this.s.sessions.push(session);
+    this.s.sessions.add(session);
     return session;
   }
 
@@ -464,27 +528,6 @@ class Topology extends EventEmitter {
     const previousTopologyDescription = this.s.description;
     const previousServerDescription = this.s.description.servers.get(serverDescription.address);
 
-    // first update the TopologyDescription
-    this.s.description = this.s.description.update(serverDescription);
-    if (this.s.description.compatibilityError) {
-      this.emit('error', new MongoError(this.s.description.compatibilityError));
-      return;
-    }
-
-    // emit monitoring events for this change
-    this.emit(
-      'serverDescriptionChanged',
-      new monitoring.ServerDescriptionChangedEvent(
-        this.s.id,
-        serverDescription.address,
-        previousServerDescription,
-        this.s.description.servers.get(serverDescription.address)
-      )
-    );
-
-    // update server list from updated descriptions
-    updateServers(this, serverDescription);
-
     // Driver Sessions Spec: "Whenever a driver receives a cluster time from
     // a server it MUST compare it to the current highest seen cluster time
     // for the deployment. If the new cluster time is higher than the
@@ -496,9 +539,41 @@ class Topology extends EventEmitter {
       resolveClusterTime(this, clusterTime);
     }
 
+    // If we already know all the information contained in this updated description, then
+    // we don't need to update anything or emit SDAM events
+    if (previousServerDescription && previousServerDescription.equals(serverDescription)) {
+      return;
+    }
+
+    // first update the TopologyDescription
+    this.s.description = this.s.description.update(serverDescription);
+    if (this.s.description.compatibilityError) {
+      this.emit('error', new MongoError(this.s.description.compatibilityError));
+      return;
+    }
+
+    // emit monitoring events for this change
+    this.emit(
+      'serverDescriptionChanged',
+      new events.ServerDescriptionChangedEvent(
+        this.s.id,
+        serverDescription.address,
+        previousServerDescription,
+        this.s.description.servers.get(serverDescription.address)
+      )
+    );
+
+    // update server list from updated descriptions
+    updateServers(this, serverDescription);
+
+    // attempt to resolve any outstanding server selection attempts
+    if (this[kWaitQueue].length > 0) {
+      processWaitQueue(this);
+    }
+
     this.emit(
       'topologyDescriptionChanged',
-      new monitoring.TopologyDescriptionChangedEvent(
+      new events.TopologyDescriptionChangedEvent(
         this.s.id,
         previousTopologyDescription,
         this.s.description
@@ -592,7 +667,7 @@ class Topology extends EventEmitter {
 
     this.selectServer(readPreferenceServerSelector(readPreference), options, (err, server) => {
       if (err) {
-        callback(err, null);
+        callback(err);
         return;
       }
 
@@ -653,19 +728,16 @@ class Topology extends EventEmitter {
     return new CursorClass(topology, ns, cmd, options);
   }
 
-  get clientInfo() {
-    return this.s.clientInfo;
+  get clientMetadata() {
+    return this.s.options.metadata;
   }
 
-  // Legacy methods for compat with old topology types
   isConnected() {
-    // console.log('not implemented: `isConnected`');
-    return true;
+    return this.s.state === STATE_CONNECTED;
   }
 
   isDestroyed() {
-    // console.log('not implemented: `isDestroyed`');
-    return false;
+    return this.s.state === STATE_CLOSED;
   }
 
   unref() {
@@ -719,16 +791,20 @@ function isWriteCommand(command) {
  *
  * @param {Server} server
  */
-function destroyServer(server, topology, callback) {
+function destroyServer(server, topology, options, callback) {
+  options = options || {};
   LOCAL_SERVER_EVENTS.forEach(event => server.removeAllListeners(event));
 
-  server.destroy(() => {
+  server.destroy(options, () => {
     topology.emit(
       'serverClosed',
-      new monitoring.ServerClosedEvent(topology.s.id, server.description.address)
+      new events.ServerClosedEvent(topology.s.id, server.description.address)
     );
 
-    if (typeof callback === 'function') callback(null, null);
+    SERVER_RELAY_EVENTS.forEach(event => server.removeAllListeners(event));
+    if (typeof callback === 'function') {
+      callback();
+    }
   });
 }
 
@@ -755,123 +831,28 @@ function randomSelection(array) {
   return array[Math.floor(Math.random() * array.length)];
 }
 
-/**
- * Selects servers using the provided selector
- *
- * @private
- * @param {Topology} topology The topology to select servers from
- * @param {function} selector The actual predicate used for selecting servers
- * @param {Number} timeout The max time we are willing wait for selection
- * @param {Number} start A high precision timestamp for the start of the selection process
- * @param {function} callback The callback used to convey errors or the resultant servers
- */
-function selectServers(topology, selector, timeout, start, callback) {
-  const duration = calculateDurationInMs(start);
-  if (duration >= timeout) {
-    return callback(new MongoTimeoutError(`Server selection timed out after ${timeout} ms`));
-  }
-
-  // ensure we are connected
-  if (!topology.s.connected) {
-    topology.connect();
-
-    // we want to make sure we're still within the requested timeout window
-    const failToConnectTimer = setTimeout(() => {
-      topology.removeListener('connect', connectHandler);
-      callback(new MongoTimeoutError('Server selection timed out waiting to connect'));
-    }, timeout - duration);
-
-    const connectHandler = () => {
-      clearTimeout(failToConnectTimer);
-      selectServers(topology, selector, timeout, process.hrtime(), callback);
-    };
-
-    topology.once('connect', connectHandler);
-    return;
-  }
-
-  // otherwise, attempt server selection
-  const serverDescriptions = Array.from(topology.description.servers.values());
-  let descriptions;
-
-  // support server selection by options with readPreference
-  if (typeof selector === 'object') {
-    const readPreference = selector.readPreference
-      ? selector.readPreference
-      : ReadPreference.primary;
-
-    selector = readPreferenceServerSelector(readPreference);
-  }
-
-  try {
-    descriptions = selector
-      ? selector(topology.description, serverDescriptions)
-      : serverDescriptions;
-  } catch (e) {
-    return callback(e, null);
-  }
-
-  if (descriptions.length) {
-    const servers = descriptions.map(description => topology.s.servers.get(description.address));
-    return callback(null, servers);
-  }
-
-  const retrySelection = () => {
-    // clear all existing monitor timers
-    topology.s.monitorTimers.map(timer => clearTimeout(timer));
-    topology.s.monitorTimers = [];
-
-    // ensure all server monitors attempt monitoring soon
-    topology.s.servers.forEach(server => {
-      const timer = setTimeout(
-        () => server.monitor({ heartbeatFrequencyMS: topology.description.heartbeatFrequencyMS }),
-        TOPOLOGY_DEFAULTS.minHeartbeatFrequencyMS
-      );
-
-      topology.s.monitorTimers.push(timer);
-    });
-
-    const descriptionChangedHandler = () => {
-      // successful iteration, clear the check timer
-      clearTimeout(iterationTimer);
-      topology.s.iterationTimers.splice(timerIndex, 1);
-
-      if (topology.description.error) {
-        callback(topology.description.error, null);
-        return;
-      }
-
-      // topology description has changed due to monitoring, reattempt server selection
-      selectServers(topology, selector, timeout, start, callback);
-    };
-
-    const iterationTimer = setTimeout(() => {
-      topology.removeListener('topologyDescriptionChanged', descriptionChangedHandler);
-      callback(new MongoTimeoutError(`Server selection timed out after ${timeout} ms`));
-    }, timeout - duration);
-
-    // track this timer in case we need to clean it up outside this loop
-    const timerIndex = topology.s.iterationTimers.push(iterationTimer);
-
-    topology.once('topologyDescriptionChanged', descriptionChangedHandler);
-  };
-
-  retrySelection();
-}
-
-function createAndConnectServer(topology, serverDescription) {
+function createAndConnectServer(topology, serverDescription, connectDelay) {
   topology.emit(
     'serverOpening',
-    new monitoring.ServerOpeningEvent(topology.s.id, serverDescription.address)
+    new events.ServerOpeningEvent(topology.s.id, serverDescription.address)
   );
 
   const server = new Server(serverDescription, topology.s.options, topology);
   relayEvents(server, topology, SERVER_RELAY_EVENTS);
 
-  server.once('connect', serverConnectEventHandler(server, topology));
   server.on('descriptionReceived', topology.serverUpdateHandler.bind(topology));
   server.on('error', serverErrorEventHandler(server, topology));
-  server.on('close', () => topology.emit('close', server));
+
+  if (connectDelay) {
+    const connectTimer = setTimeout(() => {
+      clearAndRemoveTimerFrom(connectTimer, topology.s.connectionTimers);
+      server.connect();
+    }, connectDelay);
+
+    topology.s.connectionTimers.add(connectTimer);
+    return server;
+  }
+
   server.connect();
   return server;
 }
@@ -921,28 +902,18 @@ function updateServers(topology, incomingServerDescription) {
   }
 }
 
-function serverConnectEventHandler(server, topology) {
-  return function(/* isMaster, err */) {
-    server.monitor({
-      initial: true,
-      heartbeatFrequencyMS: topology.description.heartbeatFrequencyMS
-    });
-  };
-}
-
 function serverErrorEventHandler(server, topology) {
   return function(err) {
-    topology.emit(
-      'serverClosed',
-      new monitoring.ServerClosedEvent(topology.s.id, server.description.address)
-    );
-
-    if (isSDAMUnrecoverableError(err, server)) {
-      resetServerState(server, err, { clearPool: true });
+    if (topology.s.state === STATE_CLOSING || topology.s.state === STATE_CLOSED) {
       return;
     }
 
-    resetServerState(server, err);
+    if (maxWireVersion(server) >= 8 && !isNodeShuttingDownError(err)) {
+      resetServerState(server, err);
+      return;
+    }
+
+    resetServerState(server, err, { clearPool: true });
   };
 }
 
@@ -1011,20 +982,16 @@ function executeWriteOperation(args, options, callback) {
 function resetServerState(server, error, options) {
   options = Object.assign({}, { clearPool: false }, options);
 
-  function resetState() {
-    server.emit(
-      'descriptionReceived',
-      new ServerDescription(server.description.address, null, { error })
-    );
-    server.monitor();
-  }
-
   if (options.clearPool && server.s.pool) {
-    server.s.pool.reset(() => resetState());
-    return;
+    server.s.pool.clear();
   }
 
-  resetState();
+  server.emit(
+    'descriptionReceived',
+    new ServerDescription(server.description.address, null, { error })
+  );
+
+  process.nextTick(() => server.requestCheck());
 }
 
 function translateReadPreference(options) {
@@ -1062,7 +1029,7 @@ function srvPollingHandler(topology) {
 
     topology.emit(
       'topologyDescriptionChanged',
-      new monitoring.TopologyDescriptionChangedEvent(
+      new events.TopologyDescriptionChangedEvent(
         topology.s.id,
         previousTopologyDescription,
         topology.s.description
@@ -1071,6 +1038,64 @@ function srvPollingHandler(topology) {
   };
 }
 
+function drainWaitQueue(queue, err) {
+  while (queue.length) {
+    const waitQueueMember = queue.pop();
+    clearTimeout(waitQueueMember.timer);
+    if (!waitQueueMember[kCancelled]) {
+      waitQueueMember.callback(err);
+    }
+  }
+}
+
+function processWaitQueue(topology) {
+  if (topology.s.state === STATE_CLOSED) {
+    drainWaitQueue(topology[kWaitQueue], new MongoError('Topology is closed, please connect'));
+    return;
+  }
+
+  const isSharded = topology.description.type === TopologyType.Sharded;
+  const serverDescriptions = Array.from(topology.description.servers.values());
+  for (let i = 0; i < topology[kWaitQueue].length; ++i) {
+    const waitQueueMember = topology[kWaitQueue].shift();
+    if (waitQueueMember[kCancelled]) {
+      continue;
+    }
+
+    let selectedDescriptions;
+    try {
+      const serverSelector = waitQueueMember.serverSelector;
+      selectedDescriptions = serverSelector
+        ? serverSelector(topology.description, serverDescriptions)
+        : serverDescriptions;
+    } catch (e) {
+      clearTimeout(waitQueueMember.timer);
+      waitQueueMember.callback(e);
+      break;
+    }
+
+    if (selectedDescriptions.length === 0) {
+      topology[kWaitQueue].push(waitQueueMember);
+      break;
+    }
+
+    const selectedServerDescription = randomSelection(selectedDescriptions);
+    const selectedServer = topology.s.servers.get(selectedServerDescription.address);
+    const transaction = waitQueueMember.transaction;
+    if (isSharded && transaction && transaction.isActive) {
+      transaction.pinServer(selectedServer);
+    }
+
+    clearTimeout(waitQueueMember.timer);
+    waitQueueMember.callback(undefined, selectedServer);
+  }
+
+  if (topology[kWaitQueue].length > 0) {
+    // ensure all server monitors attempt monitoring soon
+    topology.s.servers.forEach(server => process.nextTick(() => server.requestCheck()));
+  }
+}
+
 /**
  * A server opening SDAM monitoring event
  *
@@ -1155,4 +1180,6 @@ function srvPollingHandler(topology) {
  * @type {object}
  */
 
-module.exports = Topology;
+module.exports = {
+  Topology
+};
