@@ -3,6 +3,7 @@
 const Aspect = require('./operation').Aspect;
 const defineAspects = require('./operation').defineAspects;
 const OperationBase = require('./operation').OperationBase;
+const NativeTopology = require('../topologies/native_topology');
 
 class CloseOperation extends OperationBase {
   constructor(client, force) {
@@ -16,8 +17,11 @@ class CloseOperation extends OperationBase {
     const force = this.force;
     const completeClose = err => {
       client.emit('close', client);
-      for (const item of client.s.dbCache) {
-        item[1].emit('close', client);
+
+      if (!(client.topology instanceof NativeTopology)) {
+        for (const item of client.s.dbCache) {
+          item[1].emit('close', client);
+        }
       }
 
       client.removeAllListeners('close');
